//-----------------------------------------------------------------------------
// Copyright (c) 2000 Andreas Jnsson
//-----------------------------------------------------------------------------
#include "d3dapp.h"
#include <stdio.h>

//-----------------------------------------------------------------------------
// CD3DApp()
//-----------------------------------------------------------------------------
CD3DApp::CD3DApp()
{
	m_hInst = NULL;
    m_hWnd  = NULL;

	m_bReady      = false;
	m_bActive     = false;

	m_pD3DX   = 0;
	m_pD3DDev = 0;
	m_pDD     = 0;

	// Default settings
    m_szWindowTitle = "Direct3D Application";
	m_bFullscreen  = false;
	m_nWidth       = 640;
	m_nHeight      = 480;
	m_nColorBits   = D3DX_DEFAULT; // Use the same as Windows
	m_nAlphaBits   = 0;            // We don't want an alpha buffer
	m_nDepthBits   = D3DX_DEFAULT; // Use the best available
	m_nStencilBits = 0;            // We don't want a stencil buffer
}





//-----------------------------------------------------------------------------
// Create()
//-----------------------------------------------------------------------------
HRESULT CD3DApp::Create(HINSTANCE hInst, char *szCmdLine)
{
	m_hInst = hInst;

	// Create the window
	HRESULT hr;
	if( FAILED( hr = InitializeWindow() ) )
		return hr;

	// Initialize Direct3D
	if( FAILED( hr = InitializeD3D() ) )
	{
		DestroyWindow(m_hWnd);
		return hr;
	}

	// Initialize the rest
	if( FAILED( hr = Initialize(szCmdLine) ) )
	{
		DestroyWindow(m_hWnd);
		return hr;
	}

	// Start the timers
	m_AppTimer.Start();
	m_SceneTimer.Start();

	m_bReady = true;

	return S_OK;
}


//-----------------------------------------------------------------------------
// InitializeWindow()
//-----------------------------------------------------------------------------
HRESULT CD3DApp::InitializeWindow()
{
    // Register the window class
    WNDCLASS WndClass;
	WndClass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	WndClass.lpfnWndProc   = (WNDPROC)WndProc;
	WndClass.cbClsExtra    = 0;
	WndClass.cbWndExtra    = 0;
	WndClass.hInstance     = m_hInst;
	WndClass.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
	WndClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	WndClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH); 
	WndClass.lpszMenuName  = NULL;
	WndClass.lpszClassName = "Application Window";

    if( !RegisterClass( &WndClass ) )
	{
		ReportError(E_FAIL, __FILE__, __LINE__);
		return E_FAIL;
	}

    // Setup our window parameters
	int PosX, PosY, Width, Height;
	DWORD ExStyle, Style;
	if( m_bFullscreen )
	{
		ExStyle = WS_EX_TOPMOST; // This lets our window cover the whole screen
		Style   = WS_POPUP;      // We don't need caption and system 
		                         // menu when in fullscreen mode

		// Set the window to cover the whole screen
		PosX   = 0;
		PosY   = 0;
		Width  = GetSystemMetrics(SM_CXSCREEN);
		Height = GetSystemMetrics(SM_CYSCREEN);
	}
	else
	{
		ExStyle = 0;
		Style   = WS_OVERLAPPEDWINDOW; // This is a normal window

		// Make sure the client area is the size we want
		RECT Rect = {0,0,m_nWidth,m_nHeight};
		AdjustWindowRect(&Rect, Style, FALSE);

		PosX   = CW_USEDEFAULT; // Let Windows position our window
		PosY   = CW_USEDEFAULT;
		Width  = Rect.right - Rect.left;
		Height = Rect.bottom - Rect.top;
	}


    // Create the render window
	// Set lpParam to point to our application so that the window procedure
	// knows which class to call when handling window messages
    m_hWnd = CreateWindowEx( ExStyle, "Application Window", 
		                     m_szWindowTitle, Style, PosX, PosY,
						     Width,Height,NULL,NULL,m_hInst,this ); 
	if( m_hWnd == NULL )
	{
		ReportError(E_FAIL, __FILE__, __LINE__);
		return E_FAIL;
	}

	// Show the window
	ShowWindow( m_hWnd, SW_SHOW );
    UpdateWindow( m_hWnd );

	return S_OK;
}






//-----------------------------------------------------------------------------
// Run()
// Message-processing loop. 
//-----------------------------------------------------------------------------
int CD3DApp::Run()
{
	if( !m_bReady )
		return E_FAIL;

    BOOL bGotMsg;
    MSG  Msg;
    
    // Now we're ready to recieve and process Windows messages.
	PeekMessage( &Msg, NULL, 0U, 0U, PM_NOREMOVE ); // Make sure we shouldn't quit
	while( WM_QUIT != Msg.message )
    {
        // Use PeekMessage() if the app is active, so we can use idle time to
        // do the real work. Else, use GetMessage() to avoid eating CPU time.
        if( m_bActive )
            bGotMsg = PeekMessage( &Msg, NULL, 0U, 0U, PM_REMOVE );
        else
            bGotMsg = GetMessage( &Msg, NULL, 0U, 0U );
		
        if( bGotMsg )
        {
            // Translate and dispatch the message
            TranslateMessage( &Msg );
            DispatchMessage( &Msg );
        }
        else
        {
            // Let the application work during idle time
			if( m_bReady && m_bActive )
			{
				HRESULT hr;
				if( FAILED( hr = IdleAction() ) )
					DestroyWindow(m_hWnd);
			}
        }
    }

    return Msg.wParam;
}






//-----------------------------------------------------------------------------
// WndProc()
// Static msg handler which passes messages to the application class.
//-----------------------------------------------------------------------------
LRESULT CALLBACK CD3DApp::WndProc( HWND hWnd, UINT uMsg, 
								   WPARAM wParam, LPARAM lParam )
{
	CD3DApp *pApp = NULL;

	if( uMsg == WM_CREATE )
	{
		// Extract the window pointer parameter.
		CREATESTRUCT *pCS = (CREATESTRUCT *)lParam;
		pApp = (CD3DApp *)pCS->lpCreateParams;

		if( pApp )
		{
			// Save it in the window handle.
			SetWindowLong(hWnd, GWL_USERDATA, (LONG)pApp);
		}
	}
	else
	{
		// Extract the object from the window handle.
		pApp = (CD3DApp *)GetWindowLong(hWnd, GWL_USERDATA);
	}

	if( pApp )
	{
		// Let the object handle it.
		return pApp->MsgProc(hWnd, uMsg, wParam, lParam);
	}
	else
	{
		// Just do the default if the window is not created yet.
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
}







//-----------------------------------------------------------------------------
// MsgProc()
// Message handling function.
//-----------------------------------------------------------------------------
LRESULT CD3DApp::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    switch( uMsg )
    {
	case WM_ACTIVATE:
		if( LOWORD(wParam) == WA_INACTIVE || HIWORD(wParam) )
			Pause(true);
		else
			Pause(false);
		return 0;

    case WM_SIZE:
        // Check to see if the window is being hidden.
        if( SIZE_MAXHIDE == wParam || SIZE_MINIMIZED == wParam )
            Pause(true);
		else
			Pause(false);
		
        // A new window size will require new render buffer
        // sizes, so the 3D structures must be changed accordingly.
		if( m_bReady )
		{
			if( FAILED(m_pD3DX->Resize(LOWORD(lParam), HIWORD(lParam))) )
				m_bReady = false;
        }
		return 0;

	case WM_ENTERSIZEMOVE:
		// We should pause here as Windows won't allow us to 
		// do anything while the user moves or resizes the window
		Pause(true);
		return 0;

	case WM_EXITSIZEMOVE:
		Pause(false);
		return 0;

    case WM_SETCURSOR:
        // Prevent a cursor in fullscreen mode
        if( m_bActive && m_bReady && m_bFullscreen )
        {
            SetCursor(NULL);
            return 1;
        }
        break;

    case WM_SYSCOMMAND:
        // Prevent moving/sizing and power loss in fullscreen mode
        switch( wParam )
        {
            case SC_MOVE:
            case SC_SIZE:
            case SC_MAXIMIZE:
			case SC_SCREENSAVE:
            case SC_MONITORPOWER:
                if( m_bFullscreen )
                    return 1;
                break;
        }
        break;

	case WM_DESTROY:
		Destroy();

		// Tell the message loop to quit
		PostQuitMessage(0);
		return 0;
	}

	// Let the default window procedure handle 
	// any message that we don't care about
    return DefWindowProc( hWnd, uMsg, wParam, lParam );
}





//-----------------------------------------------------------------------------
// IdleAction()
// This is called every time the application don't have any messages to process
//-----------------------------------------------------------------------------
HRESULT CD3DApp::IdleAction()
{
	if( !m_bReady )
		return E_FAIL;


	// Update the application state
	HRESULT hr;
	if( FAILED(hr = UpdateScene()) )
		return hr;

	// Render the scene
	if( FAILED(hr = RenderScene()) )
		return hr;

	// Show the scene
    hr = m_pD3DX->UpdateFrame( 0 );
    if( hr == DDERR_SURFACELOST || hr == DDERR_SURFACEBUSY )
	{
		// Try to recover
		if( FAILED(hr = HandleModeChanges()) )
			return hr;
	}

	// Tick timers
	m_AppTimer.Frame();
	m_SceneTimer.Frame();
	
	// Update the framerate
	char Title[256];
	char OldTitle[256];
	sprintf(Title, "%s (%.3f)", m_szWindowTitle, m_AppTimer.GetFps());
	GetWindowText(m_hWnd, OldTitle, 256);


	// Check to see if anything has changed. If we don't do this  
	// we would get some serious flickering at higher frame rates.
	if( strcmp(Title, OldTitle) )
		SetWindowText(m_hWnd, Title);

	return S_OK;
}





//-----------------------------------------------------------------------------
// Pause()
// Pause the application in a way that the timers don't count while the app is 
// inactive
//-----------------------------------------------------------------------------
HRESULT CD3DApp::Pause(bool bPause)
{
	if( bPause )
	{
		m_bActive = false;
		m_AppTimer.Stop();

		// We need to remember if the scene timer was running, otherwise
		// we will start it again when the application is activated again, 
		// regardless if the timer was paused or not.
		if( m_SceneTimer.IsRunning() )
		{
			m_SceneTimer.Stop();
			m_bSceneIsRunning = true;
		}
		else 
			m_bSceneIsRunning = false;

		char Title[256];
		sprintf(Title, "%s (paused)", m_szWindowTitle);
		SetWindowText(m_hWnd, Title);
	}
	else
	{
		m_bActive = true;
		m_AppTimer.Continue();
		if( m_bSceneIsRunning )
			m_SceneTimer.Continue();
	}

	return S_OK;
}





//-----------------------------------------------------------------------------
// Destroy()
//-----------------------------------------------------------------------------
HRESULT CD3DApp::Destroy()
{
	m_bReady = false;

    // Terminate the application
	Terminate();

	// Terminate Direct3D
	TerminateD3D();

	return S_OK;
}





//-----------------------------------------------------------------------------
// HandleModeChanges()
// Call this whenever pD3DX->UpdateFrame() returns DDERR_SURFACELOST or 
// DDERR_SURFACEBUSY. The function takes care of cases where another program 
// unexpectedly changes window mode or otherwise interupts our connection with 
// the screen.
//-----------------------------------------------------------------------------
HRESULT CD3DApp::HandleModeChanges()
{
    HRESULT hr;

	// Check if we can continue with our rendering
    hr = m_pDD->TestCooperativeLevel();
    if( SUCCEEDED( hr ) )
    {
        // This means that another program borrowed the screen for a while, but
		// was nice enough to restore it to our original mode. We have still 
		// lost our memoryallocations for our surfaces though, so we need to 
		// restore them.
        if( FAILED( m_pDD->RestoreAllSurfaces() ) )
		{
			ReportError(hr, __FILE__, __LINE__);
            return hr;
		}
    }
    else 
    {
		// Another application has stolen the screen from us.
		ReportError(hr, __FILE__, __LINE__);
		return hr;
	}

	return S_OK;
}


//-----------------------------------------------------------------------------
// InitializeD3D()
//-----------------------------------------------------------------------------
HRESULT CD3DApp::InitializeD3D()
{
	// Initialize D3DX
	HRESULT hr;
    if( FAILED(hr = D3DXInitialize()) )
	{
		ReportError(hr, __FILE__, __LINE__);
        return hr;
	}

	// Create the D3DX context
	// D3DX creates everything we need here
	DWORD Flags = m_bFullscreen ? D3DX_CONTEXT_FULLSCREEN : 0;
	hr = D3DXCreateContextEx( D3DX_DEFAULT, // Use the best acceleration
		                      Flags,
							  m_hWnd,
							  NULL,         // We don't use a parent window
							  m_nColorBits,
							  m_nAlphaBits,
							  m_nDepthBits,
							  m_nStencilBits,
							  D3DX_DEFAULT, // Use one backbuffer
							  m_nWidth,
							  m_nHeight,
							  D3DX_DEFAULT, // Let D3DX choose refreshrate
							  &m_pD3DX );
    if( FAILED(hr) )
	{
		ReportError(hr, __FILE__, __LINE__);
		return hr;
	}

	// Get a reference to Direct3DDevice
	m_pD3DDev = m_pD3DX->GetD3DDevice();
    if( m_pD3DDev == NULL )
	{
		ReportError(E_FAIL, __FILE__, __LINE__);
		return E_FAIL;
	}

	// Get a reference to DirectDraw
    m_pDD = m_pD3DX->GetDD();
    if( m_pDD == NULL )
	{
		ReportError(E_FAIL, __FILE__, __LINE__);
		return E_FAIL;
	}

	return S_OK;
}


//-----------------------------------------------------------------------------
// TerminateD3D()
//-----------------------------------------------------------------------------
HRESULT CD3DApp::TerminateD3D()
{
	// Release our DirectX objects
	if( m_pDD )
	{
		m_pDD->Release();
		m_pDD = 0;
	}

	if( m_pD3DDev )
	{
		m_pD3DDev->Release();
		m_pD3DDev = 0;
	}

	if( m_pD3DX )
	{
		m_pD3DX->Release();
		m_pD3DX = 0;
	}

	// Let D3DX do some final uninitializations
	D3DXUninitialize();

	return S_OK;
}



//-----------------------------------------------------------------------------
// ReportError()
//-----------------------------------------------------------------------------
void CD3DApp::ReportError(HRESULT hr, char *szFile, int nLine)
{
	char szError[256];
	D3DXGetErrorString(hr, 256, szError);

	char szMessage[1000];
	sprintf(szMessage, "error: \t%s\nfile: \t%s\nline: \t%d\n", szError, szFile, nLine);

	// Show the error message to the user
	if( !m_bFullscreen )
		MessageBox(NULL, szMessage, "Error", MB_OK|MB_ICONERROR|MB_SYSTEMMODAL);

	// Show the error message to the developer
	OutputDebugString(szMessage);
}


//-----------------------------------------------------------------------------
// Initialize()
// Initialize our application specific objects here
//-----------------------------------------------------------------------------
HRESULT CD3DApp::Initialize(char *szCmdLine)
{
	return S_OK;
}



//-----------------------------------------------------------------------------
// Terminate()
// Application specific clean up code goes here
//-----------------------------------------------------------------------------
HRESULT CD3DApp::Terminate()
{
	return S_OK;
}


//-----------------------------------------------------------------------------
// UpdateScene()
// Update the your scene here
//-----------------------------------------------------------------------------
HRESULT CD3DApp::UpdateScene()
{
	return S_OK;
}


//-----------------------------------------------------------------------------
// RenderScene()
// This is called to render the scene
//-----------------------------------------------------------------------------
HRESULT CD3DApp::RenderScene()
{
	// Clear the buffers
	m_pD3DX->Clear(D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER);

	// Begin a new scene
    HRESULT hr = m_pD3DDev->BeginScene();
	if( SUCCEEDED( hr ) )
	{
		// Render here

		// The scene is finished
        hr = m_pD3DDev->EndScene();
	}

	if( FAILED(hr) )
	{
		ReportError(hr, __FILE__, __LINE__);
		return hr;
	}

	return S_OK;
}
